import numpy as np
import time
from tqdm import tqdm
import random
import math
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import gym
from collections import deque, namedtuple
from torch.utils.tensorboard import SummaryWriter
Transition = namedtuple('Transition', ('state', 'action', 'reward', 'future_state', 'done'))

class History:
    def __init__(self):
        t = time.localtime()
        self.writer = SummaryWriter(
            f'runs/haleluja2layers-{str(t.tm_hour) + "_" + str(t.tm_min)}')  # tensorboard --logdir=runs
        self.ep_rewards = []
        self.aggr_ep_rewards = {'ep': [], 'avg': [], 'min': [], 'max': [], 'acc_landings': []}

    def update_board(self, loss, epsilon):
        self.writer.add_scalar('loss_QNET', loss, len(self.ep_rewards))
        self.writer.add_scalar('epsilon', epsilon, len(self.ep_rewards))
        self.writer.add_scalar('avg_reward', self.aggr_ep_rewards['avg'][-1], len(self.ep_rewards))
        self.writer.add_scalar('acc_landing', self.aggr_ep_rewards['acc_landings'][-1], len(self.ep_rewards))
        self.writer.add_scalar('max_reward', self.aggr_ep_rewards['max'][-1], len(self.ep_rewards))
        self.writer.add_scalar('min_reward', self.aggr_ep_rewards['min'][-1], len(self.ep_rewards))

    def update_reward(self, reward):
        self.ep_rewards.append(reward)

    def update_correct_landing(self, land_acc):
        self.aggr_ep_rewards["acc_landings"].append(land_acc)

    def print_last(self, episode):
        print(f"Episode {episode} reward: {self.ep_rewards[-1]}")

    def print_average_last(self, episode, tail):
        average_reward = np.round(sum(self.ep_rewards[-tail:]) / tail, 4)
        self.aggr_ep_rewards['ep'].append(episode)
        self.aggr_ep_rewards['avg'].append(average_reward)
        self.aggr_ep_rewards['min'].append(np.round(min(self.ep_rewards[-tail:]), 4))
        self.aggr_ep_rewards['max'].append(np.round(max(self.ep_rewards[-tail:]), 4))
        print(f"Episode: {episode} | Reward avg: {average_reward} min: "
              f"{self.aggr_ep_rewards['min'][-1]} max: {self.aggr_ep_rewards['max'][-1]}"
              f" landing_accuracy: {np.round(self.aggr_ep_rewards['acc_landings'][-1] * 100, 2)}% (over last {tail} episodes)")


class DQN(nn.Module):
    def __init__(self, input_size, hidden_dims, output_size):
        super(DQN, self).__init__()
        self.input_size = input_size
        self.output_size = output_size
        self.n_dims = len(hidden_dims)
        self.hid_act = F.relu

        self.layers = []
        for h in hidden_dims:
            self.layers.append(torch.nn.Linear(input_size, h))
            input_size = h
        self.layers.append(torch.nn.Linear(input_size, output_size))
        self.layers = nn.ModuleList(self.layers)

    def forward(self, x):
        for i, layer in enumerate(self.layers):
            if self.n_dims != i:
                x = self.hid_act(layer(x))
            else:
                x = layer(x)
        return x



class DQNAgent:
    class ReplayMemory(object):

        def __init__(self, capacity):
            self.memory = deque([], maxlen=capacity)

        def push(self, *args):
            """Save a transition"""
            self.memory.append(Transition(*args))

        def sample(self, batch_size):
            return random.sample(self.memory, batch_size)

        def __len__(self):
            return len(self.memory)

    def __init__(self, env, C, hidden_layers, replay_memory_size, optimizer, loss, device, lr):
        self.env = env
        self.obs_space_size = env.observation_space.shape[0]  # (6 continuous vars and 2 boolean)

        self.cost_function = loss
        self.device = device
        self._init_networks(hidden_layers)
        self.replay_memory = self.ReplayMemory(replay_memory_size)
        self.losses_memory = deque(maxlen=150)
        self.C = C  # (target network update step constant)
        self.epsilon = None
        self.optimizer = optimizer(self.Q_net.parameters(), lr)

    def _init_networks(self, hidden_layers):
        self.Q_net = DQN(input_size=self.obs_space_size, hidden_dims=hidden_layers,
                         output_size=self.env.action_space.n)
        self.target_Q_net = DQN(input_size=self.obs_space_size, hidden_dims=hidden_layers,
                                output_size=self.env.action_space.n)
        self.Q_net.to(self.device)
        self.target_Q_net.to(self.device)
        self.target_Q_net.load_state_dict(self.Q_net.state_dict())
        self.target_Q_net.eval()

    def save_model(self, episode):
        t = time.localtime()
        torch.save(self.Q_net.state_dict(), f'models/ep_{episode}_time{str(t.tm_hour) + "_" + str(t.tm_min)}.pth')

    def _is_correctlanding(self, state):
        '''
        Is it standing on both legs near the platoue (0, 0) ? thats usualy in interval the <-0.7, 0.7>
        :return: boolean
        '''
        state = state[0]
        if state[-1].item() == 1 and state[-2].item() == 1 and (-0.1 <= state[0].item() <= 0.1) and (-0.1 <= state[1].item() <= 0.1):
            return True
        return False

    def _make_tensor(self, x):
        return torch.tensor(x).to(self.device)

    def _store_transition(self, transition):
        '''
        :param transition: includes (state, action, reward, future_state, done)
        if the state is terminal done == true
        :return:
        '''
        self.replay_memory.push(*transition)

    def _transition_get_states(self, mini_batch):
        states = self._make_tensor([transition[0] for transition in mini_batch])
        return states

    def _transition_get_future_states(self, mini_batch):
        states = self._make_tensor([transition[3] for transition in mini_batch])
        return states

    def train_episode(self, batch_size, gamma):
        if len(self.replay_memory) < batch_size:
            return

        transitions = self.replay_memory.sample(batch_size)
        mini_batch = Transition(*zip(*transitions))
        non_final_mask = torch.tensor(tuple(map(lambda d: not d, mini_batch.done)), device=self.device, dtype=torch.bool)
        non_final_next_states = torch.cat([s for s, d in zip(mini_batch.future_state, mini_batch.done) if not d])

        state_batch = torch.cat(mini_batch.state)
        action_batch = torch.cat(mini_batch.action)
        reward_batch = torch.cat(mini_batch.reward)
        state_action_values = self.Q_net(state_batch).gather(1, action_batch)

        next_state_values = torch.zeros(batch_size, device=device)
        next_state_values[non_final_mask] = self.target_Q_net(non_final_next_states).max(1)[0].detach()
        expected_state_action_values = (next_state_values * gamma) + reward_batch

        loss = self.cost_function(state_action_values, expected_state_action_values.unsqueeze(1))
        self.losses_memory.append(loss.item())
        self.Q_net.zero_grad()
        loss.backward()
        for param in self.Q_net.parameters():
            param.grad.data.clamp_(-1, 1)
        self.optimizer.step()

    def select_action(self, state):
        if np.random.random() > self.epsilon:
            with torch.no_grad():
                best_action = torch.tensor([[self.Q_net(state).argmax()]], device=self.device)
                return best_action
        else:
            random_action = torch.tensor([[random.randrange(env.action_space.n)]], device=self.device)
            return random_action


    def train(self, episodes, batch_size, gamma, epsilon_start, epsilon_end, eps_decay, render=True, verbose=True, verbose_step=100):
        history = History()
        episodes_correct = 0
        steps = 0
        decay_steps = 0
        for episode in tqdm(range(1, episodes + 1), ascii=True, unit='episodes'):
            episode_reward = 0
            current_state = self.env.reset()
            current_state = torch.tensor([current_state]).to(self.device)
            done = False
            self.epsilon = epsilon_end + (epsilon_start - epsilon_end) * math.exp(-1. * decay_steps / eps_decay)
            decay_steps += 1
            while not done:
                action = self.select_action(current_state)
                new_state, reward, done, _ = self.env.step(action.item())
                new_state = torch.tensor([new_state]).to(self.device)
                reward = torch.tensor([reward], dtype=torch.float).to(self.device)
                episode_reward += reward.item()
                self.replay_memory.push(current_state, action, reward, new_state, done)
                current_state = new_state
                self.train_episode(batch_size, gamma)
                steps += 1

                if render and (episode - 1) % (verbose_step * 10) == 0:
                    self.env.render()
                    time.sleep(0.01)
                if steps % self.C == 0:
                    self.target_Q_net.load_state_dict(self.Q_net.state_dict())
                    steps = 0
                if done and self._is_correctlanding(new_state):
                    episodes_correct += 1
            episodes_correct = self.update_history(history, episode_reward, episode, verbose_step, verbose,
                                                   episodes_correct)
            self.check_point(episode)

    def check_point(self, episode):
        if episode % 10 == 0:
            self.save_model(episode)

    def update_history(self, history, episode_reward, epis, verbose_step, verbose, episodes_correct):
        history.update_reward(episode_reward)
        if epis % verbose_step == 0:
            history.update_correct_landing(episodes_correct / verbose_step)
            if verbose:
                history.print_average_last(epis, tail=verbose_step)
                la = sum(self.losses_memory) / len(self.losses_memory)
                history.update_board(la, self.epsilon)
            else:
                history.print_last(epis)
            episodes_correct = 0
        return episodes_correct

    def load_Q_net(self, file):
        self.Q_net.load_state_dict(torch.load(file))
        self.target_Q_net.load_state_dict(self.Q_net.state_dict())
        self.target_Q_net.eval()

    def evaluate(self, episodes=10_000):
        rewards_list = []
        correct_landing = 0
        self.epsilon = 0
        self.Q_net.eval()
        for i in range(episodes):
            current_state = self.env.reset()
            current_state = torch.tensor([current_state]).to(self.device)
            done = False
            episode_reward = 0
            while not done:
                action = self.select_action(current_state)
                new_state, reward, done, _ = self.env.step(action.item())
                new_state = torch.tensor([new_state]).to(self.device)
                current_state = new_state
                episode_reward += reward
                if done and self._is_correctlanding(new_state):
                    correct_landing += 1

            rewards_list.append(episode_reward)
            print(f'landings: {correct_landing}, episode: {i}, reward: {episode_reward}')
        return rewards_list, correct_landing

    def run_episode(self):
        self.Q_net.eval()
        current_state = self.env.reset()
        current_state = torch.tensor([current_state]).to(self.device)
        done = False
        self.epsilon = 0
        episode_reward = 0
        while not done:
            action = self.select_action(current_state)
            new_state, reward, done, _ = self.env.step(action.item())
            episode_reward += reward
            current_state = torch.tensor([new_state]).to(self.device)
            self.env.render()
            time.sleep(0.0125)
            print(episode_reward)


if __name__ == '__main__':
    if torch.cuda.is_available():
        if torch.cuda.device_count() == 2:
            dev = "cuda:1"
        else:
            dev = "cuda:0"
        print('Running on GPU {:}'.format(dev))
    else:
        dev = "cpu"
        print('Running on CPU')

    device = torch.device(dev)
    env = gym.make("LunarLander-v2")

    agent = DQNAgent(env, C=2000, hidden_layers=[64, 64], replay_memory_size=10000, optimizer=optim.Adam, loss=nn.SmoothL1Loss(), device=device, lr=0.00005)
    agent.load_Q_net("for_animation/880.pth")
    agent.run_episode()
    # , , , , , , , ,
    #rewards_list, correct_landing = agent.evaluate(episodes=10_000)
    #agent.train(episodes=2000, batch_size=128, gamma=0.99, epsilon_start=0.1, epsilon_end=0.1, eps_decay=1000, render=True, verbose=True, verbose_step=25)
