import numpy as np
import time
from utils import History

class QLunarAgent:
    def __init__(self, env, alpha, gamma, epsilon_start, epsilon_end, mode, bins):
        self.env = env
        self.alpha = alpha
        self.gamma = gamma
        self.eps_s = epsilon_start
        self.eps_e = epsilon_end
        # Observed min/max (x,y,velocity_x/y,angle,angle_velocity values of space ship (observed with histograms)
        self.environment_border = {'min_x': -1, 'max_x': 1, 'min_y': -1, 'max_y': 1.7,
                                   'min_vx': -2, 'max_vx': 2, 'min_vy': -2 , 'max_vy': 0.6,
                                   'min_ang': -2, 'max_ang': 2, 'min_vang': -1.2, 'max_vang': 1.2}
        self.observation_space_low = None
        self.observation_space_high = None
        self.window_size = None
        self.state_idx = None
        self.mode = mode
        self.bins = bins
        self.Q_table = None
        self.initialize_Qtable()

    def load_qtable(self, file_name):
        self.Q_table = np.load(file_name)

    def initialize_Qtable(self):
        '''
        state space consists of 4 unbounded Real values and 2 bools of
        (pos_x, pos_y, vel_x, vel_y, ang, ang_vel, left_leg_contact, right_leg_contact)
        self.mode: mode = 1 makes Q_table that contain states (pos_x, pos_y),
                   mode = 2 makes Q_table that contain states (pos_x, pos_y, left_leg_contact, right_leg_contact)
        self.bins: number of bins no_bins of {x:N1, y:N2, ...}
        :return: random Q_table
        '''
        min_x, min_y = self.environment_border['min_x'], self.environment_border['min_y']
        max_x, max_y = self.environment_border['max_x'], self.environment_border['max_y']
        if self.mode == 1:
            self.observation_space_low = np.array([min_x, min_y])
            self.observation_space_high = np.array([max_x, max_y])
            discrete_state_map_sizes = [self.bins['x'], self.bins['y']]
            self.window_size = (self.observation_space_high - self.observation_space_low)/discrete_state_map_sizes
            self.state_idx = [0, 1]
            self.Q_table = np.zeros((self.bins['x'], self.bins['y'], self.env.action_space.n))
        if self.mode == 2:
            self.observation_space_low = np.array([min_x, min_y, 0, 0])
            self.observation_space_high = np.array([max_x, max_y, 1, 1])
            discrete_state_map_sizes = [self.bins['x'], self.bins['y'], self.bins['left_leg']-1, self.bins['right_leg']-1]
            self.window_size = (self.observation_space_high - self.observation_space_low) / discrete_state_map_sizes
            self.state_idx = [0, 1, 6, 7]
            self.Q_table = np.zeros((self.bins['x'], self.bins['y'], self.bins['left_leg'], self.bins['right_leg'], self.env.action_space.n))
        if self.mode == 3:
            max_vx, max_vy = self.environment_border['max_vx'], self.environment_border['max_vy']
            min_vx, min_vy = self.environment_border['min_vx'], self.environment_border['min_vy']
            min_ang, min_vang = self.environment_border['min_ang'], self.environment_border['min_vang']
            max_ang, max_vang = self.environment_border['max_ang'], self.environment_border['max_vang']
            self.observation_space_low = np.array([min_x, min_y, min_vx, min_vy, min_ang, min_vang, 0, 0])
            self.observation_space_high = np.array([max_x, max_y, max_vx, max_vy, max_ang, max_vang, 1, 1])
            discrete_state_map_sizes = [self.bins['x'], self.bins['y'], self.bins['vx'], self.bins['vy'],
                                        self.bins['a'], self.bins['va'], self.bins['left_leg'] - 1, self.bins['right_leg'] - 1]
            self.window_size = (self.observation_space_high - self.observation_space_low) / discrete_state_map_sizes
            self.state_idx = [0, 1, 2, 3, 4, 5, 6, 7]
            self.Q_table = np.zeros((self.bins['x'], self.bins['y'], self.bins['vx'], self.bins['vy'],
                                     self.bins['a'], self.bins['va'], self.bins['left_leg'],
                                     self.bins['right_leg'], self.env.action_space.n))

    def _normalize_discrete(self, discrete_s):
        for s_i in range(len(discrete_s)):
            discrete_s[s_i] = max(0, discrete_s[s_i])
            max_s_i = self.Q_table.shape[s_i] - 1
            discrete_s[s_i] = min(max_s_i, discrete_s[s_i])
        return tuple(discrete_s)

    def discrete_state(self, state):
        mode_state = state[self.state_idx]
        discrete_s = (mode_state - self.observation_space_low) / self.window_size
        discrete_s = list(discrete_s.astype(np.int))
        discrete_s = self._normalize_discrete(discrete_s)
        return discrete_s

    def _is_correctlanding(self, state):
        '''
        Is it standing on both legs near the platoue (0, 0) ? thats usualy in interval the <-0.7, 0.7>
        :return: boolean
        '''
        if state[-1] == 1 and state[-2] == 1 and (-0.1 <= state[0] <= 0.1) and (-0.1 <= state[1] <= 0.1):
            return True
        return False

    def save_qtable(self, name):
        np.save(f'tables/{name}.npy', self.Q_table)

    def train(self, episodes, render=True, verbose=True, verbose_step=100):
        history = History()
        episodes_correct = 0
        for step in range(1, episodes+1):
            r = np.max((episodes-step)/episodes, 0)
            epsilon = (self.eps_s - self.eps_e)*r + self.eps_e
            episode_reward = 0
            discrete_state = self.discrete_state(self.env.reset())
            done = False
            while not done:
                if np.random.random() > epsilon:
                    action = np.argmax(self.Q_table[discrete_state])
                else:
                    action = np.random.randint(0, self.env.action_space.n)
                new_state, reward, done, _ = self.env.step(action)
                episode_reward += reward
                if render and (step-1) % verbose_step == 0:
                    self.env.render()
                    time.sleep(0.01)
                if not done:
                    new_discrete_state = self.discrete_state(new_state)
                    q_now = self.Q_table[discrete_state + (action, )]
                    q_max_future = np.max(self.Q_table[new_discrete_state])
                    q_new = q_now + self.alpha * (reward + self.gamma * q_max_future - q_now)
                    self.Q_table[discrete_state + (action,)] = q_new
                    discrete_state = new_discrete_state
                elif self._is_correctlanding(new_state):
                    episodes_correct += 1

            history.update_reward(episode_reward)
            if step % verbose_step == 0:
                self.save_qtable(f'episode_{step}')
                np.save(f'history/history_ep_rewards_episode_{step}.npy', history.ep_rewards)
                np.save(f'history/history_rewards_avg_episode_{step}.npy', history.aggr_ep_rewards['avg'])
                np.save(f'history/history_acc_landings_episode_{step}.npy', history.aggr_ep_rewards['acc_landings'])
                history.update_correct_landing(episodes_correct/verbose_step)
                if verbose:
                    history.print_average_last(step, tail=verbose_step)
                else:
                    history.print_last(step)
                episodes_correct = 0
        return history

    def evaluate(self, episodes=10_000):
        rewards_list = []
        correct_landing = 0
        for i in range(episodes):
            episode_reward = 0
            discrete_state = self.discrete_state(self.env.reset())
            done = False
            while not done:
                action = np.argmax(self.Q_table[discrete_state])
                new_state, reward, done, _ = self.env.step(action)
                episode_reward += reward
                discrete_state = self.discrete_state(new_state)
                if done and self._is_correctlanding(new_state):
                    correct_landing += 1
            rewards_list.append(episode_reward)
            print(f'landings: {correct_landing}, episode: {i}, reward: {episode_reward}')
        return rewards_list, correct_landing

    def run_episode(self):
        discrete_state = self.discrete_state(self.env.reset())
        done = False
        episode_reward = 0
        while not done:
            action = np.argmax(self.Q_table[discrete_state])
            new_state, reward, done, _ = self.env.step(action.item())
            episode_reward += reward
            discrete_state = self.discrete_state(new_state)
            self.env.render()
            time.sleep(0.0125)
            print(episode_reward)